# Copyright 2017, 2018 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import ujson as json
import subprocess

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import BgpSession, EvpnVniInfo, Description
from netq_lib.common.enums import ServiceStatus


def process_bgp_neighbors(output, svc_failed=False, **kwargs):
    """Build summary of BGP node info, such as ASN & RID
    """
    retcode = ServiceStatus.OK

    try:
        json_out = json.loads(output)
    except ValueError:
        # We treat invalid JSON as the infomation terminating
        json_out = []
        retcode = ServiceStatus.FAIL

    # first process the entries already in the DB
    m_bgp_session = BgpSession.memmodel()
    for entry in list(m_bgp_session.query.filter(hostname=HOSTNAME)):
        if entry.state == 'Fail':
            continue
        if entry.vrf == 'default':
            vrf = 'Default'
        else:
            vrf = entry.vrf
        old_bgp_entry = BgpSession()
        if (vrf not in json_out or
                entry.peer_name not in json_out[vrf]):
            if not svc_failed:
                old_bgp_entry.copy(entry)
                if old_bgp_entry.objid:
                    old_descr = Description()
                    old_descr.objtype = old_bgp_entry.__class__.__name__
                    old_descr.descrid = old_bgp_entry.objid
                    old_descr.delete()
                old_bgp_entry.delete()
            else:
                old_bgp_entry.state = 'Fail'
                old_bgp_entry.vrf = entry.vrf
                old_bgp_entry.asn = entry.asn
                old_bgp_entry.peer_name = entry.peer_name
                old_bgp_entry.save()
        else:
            # Check that ASN hasn't changed
            if json_out[vrf][entry.peer_name]['localAs'] != entry.asn:
                old_bgp_entry.copy(entry)
                if old_bgp_entry.objid:
                    old_descr = Description()
                    old_descr.objtype = old_bgp_entry.__class__.__name__
                    old_descr.descrid = old_bgp_entry.objid
                    old_descr.delete()
                old_bgp_entry.delete()

            # Check that the description hasn't been deleted
            if (entry.objid and
                    not json_out[vrf][entry.peer_name].get('nbrDesc', '')):
                descr = Description()
                descr.objtype = entry.__class__.__name__
                descr.descrid = entry.objid
                descr.delete()

    for vrf in json_out:
        vrf_id = json_out[vrf]['vrfId']
        for key in json_out[vrf]:
            if key == 'vrfName' or key == 'vrfId':
                continue

            session = json_out[vrf][key]

            bgp_nbr = BgpSession()
            # Quagga/FRR names the default VRF 'Default' whereas Linux names it
            # nothing, and in netq we call it 'default'. Fix this.
            if vrf == 'Default':
                bgp_nbr.vrf = 'default'
            else:
                bgp_nbr.vrf = vrf.encode('ascii', 'ignore')
            bgp_nbr.vrfid = vrf_id
            bgp_nbr.peer_name = key.encode('ascii', 'ignore')

            # protect against new sections that are not 'sessions'
            if session.get('bgpState') is None:
                continue
            bgp_nbr.state = session['bgpState'].encode('ascii', 'ignore')
            if bgp_nbr.state != 'Established':
                bgp_nbr.state = 'NotEstd'
            bgp_nbr.asn = int(session['localAs'])
            bgp_nbr.reason = 'N/A'
            if bgp_nbr.state == 'Established':
                bgp_nbr.up_time = float(
                    session.get('bgpTimerUpEstablishedEpoch', 0.0))
            else:
                bgp_nbr.up_time = 0.0
            bgp_nbr.peer_router_id = session['remoteRouterId'].encode('ascii',
                                                                      'ignore')
            bgp_nbr.peer_asn = session['remoteAs']
            bgp_nbr.peer_hostname = session.get('hostname', '').encode('ascii',
                                                                       'ignore')
            bgp_nbr.upd8_rx = session['messageStats']['updatesRecv']
            bgp_nbr.upd8_tx = session['messageStats']['updatesSent']
            afi_session = session['addressFamilyInfo']
            if 'lastResetDueTo' in afi_session:
                bgp_nbr.reason = (
                    afi_session['lastResetDueTo'].encode('ascii', 'ignore')
                )
                bgp_nbr.last_reset_time = \
                    float(afi_session['lastResetTimerMsecs'] / 1000)
            elif 'lastResetDueTo' in session:
                bgp_nbr.reason = (
                    session['lastResetDueTo'].encode('ascii', 'ignore')
                )
                if bgp_nbr.reason.startswith("BGP Notification"):
                    if 'lastNotificationReason' in session:
                        bgp_nbr.reason = (
                            session['lastNotificationReason'].encode('ascii',
                                                                     'ignore'))
                bgp_nbr.last_reset_time = \
                    float(session['lastResetTimerMsecs'] / 1000)
            else:
                bgp_nbr.last_reset_time = 0.0
            if 'IPv4 Unicast' in session['addressFamilyInfo']:
                bgp_nbr.ipv4_pfx_rcvd = (
                    int(afi_session['IPv4 Unicast']['acceptedPrefixCounter'])
                )
            else:
                bgp_nbr.ipv4_pfx_rcvd = 0
            if 'IPv6 Unicast' in session['addressFamilyInfo']:
                bgp_nbr.ipv6_pfx_rcvd = (
                    int(afi_session['IPv6 Unicast']['acceptedPrefixCounter'])
                )
            else:
                bgp_nbr.ipv6_pfx_rcvd = 0
            if 'L2VPN EVPN' in session['addressFamilyInfo']:
                bgp_nbr.evpn_pfx_rcvd = (
                    int(afi_session['L2VPN EVPN']['acceptedPrefixCounter']))
            else:
                bgp_nbr.evpn_pfx_rcvd = 0
            if 'connectionsDropped' in afi_session:
                bgp_nbr.conn_dropped = afi_session['connectionsDropped']
                bgp_nbr.conn_estd = afi_session['connectionsEstablished']
            elif 'connectionsDropped' in session:
                bgp_nbr.conn_dropped = session['connectionsDropped']
                bgp_nbr.conn_estd = session['connectionsEstablished']

            # BGP Capabilities
            rx_families = []
            tx_families = []
            if 'neighborCapabilities' in session:
                if 'multiprotocolExtensions' in session['neighborCapabilities']:
                    caps_dict = session['neighborCapabilities']['multiprotocolExtensions']
                    if 'IPv4 Unicast' in caps_dict:
                        if caps_dict['IPv4 Unicast'].get(
                                'advertisedAndReceived', False):
                            tx_families.append('ipv4')
                            rx_families.append('ipv4')
                        elif caps_dict['IPv4 Unicast'].get(
                                'advertised', False):
                            tx_families.append('ipv4')
                        elif caps_dict['IPv4 Unicast'].get(
                                'received', False):
                            rx_families.append('ipv4')
                    if 'IPv6 Unicast' in caps_dict:
                        if caps_dict['IPv6 Unicast'].get(
                                'advertisedAndReceived', False):
                            tx_families.append('ipv6')
                            rx_families.append('ipv6')
                        elif caps_dict['IPv6 Unicast'].get(
                                'advertised', False):
                            tx_families.append('ipv6')
                        elif caps_dict['IPv6 Unicast'].get(
                                'received', False):
                            rx_families.append('ipv6')
                    if 'L2VPN EVPN' in caps_dict:
                        if caps_dict['L2VPN EVPN'].get(
                                'advertisedAndReceived', False):
                            tx_families.append('evpn')
                            rx_families.append('evpn')
                        elif caps_dict['L2VPN EVPN'].get(
                                'advertised', False):
                            tx_families.append('evpn')
                        elif caps_dict['L2VPN EVPN'].get(
                                'received', False):
                            rx_families.append('evpn')
                bgp_nbr.tx_families = tx_families
                bgp_nbr.rx_families = rx_families

            # Save description if present
            descr = session.get('nbrDesc', '')
            if descr:
                bgp_descr = Description()
                bgp_descr.objtype = bgp_nbr.__class__.__name__
                bgp_descr.descrid = bgp_nbr.get_objid()
                bgp_descr.description = descr.encode('ascii', 'ignore')
                bgp_descr.timestamp = bgp_nbr.timestamp
                bgp_nbr.objid = bgp_descr.descrid
                bgp_descr.save()
            bgp_nbr.save()

    return retcode


def process_bgp_evpn_vni(output, svc_failed=False, **kwargs):
    '''Process EVPN VNI command output'''
    retcode = ServiceStatus.OK

    try:
        json_out = json.loads(output)
    except ValueError:
        # We treat invalid JSON as the infomation terminating
        json_out = []
        retcode = ServiceStatus.FAIL

    # first process the entries already in the DB
    m_evpn_info = EvpnVniInfo.memmodel()
    no_entry = True
    for entry in list(m_evpn_info.query.filter(hostname=HOSTNAME)):
        no_entry = False
        if retcode == ServiceStatus.FAIL:
            new_entry = EvpnVniInfo()
            new_entry.copy(entry)
            new_entry.delete()
            continue

        vni = str(entry.vni)
        if vni == 0:
            continue
        if vni not in json_out:
            new_entry = EvpnVniInfo()
            new_entry.copy(entry)
            new_entry.delete()
            continue

        # Check that the RD hasn't changed (happens when a VXLAN dev is
        # deleted and added back
        if json_out[vni]['rd'].encode('ascii', 'ignore') != entry.rd:
            new_entry = EvpnVniInfo()
            new_entry.copy(entry)
            new_entry.delete()

    if not json_out and no_entry:
        # There was no JSON output, but thats because there's no EVPN either
        retcode = ServiceStatus.OK

    if json_out:
        try:
            adv_all_vni = (
                json_out['advertiseAllVnis'].encode('ascii',
                                                'ignore') == 'Enabled')
            adv_gw_ip = json_out['advertiseGatewayMacip'].encode('ascii',
                                                             'ignore')
        except ValueError:
            # assume defaults
            adv_all_vni = 'Disabled'
            adv_gw_ip = 'Disabed'

    for key in json_out:
        # we only care about the vni info which is a dict
        # ignore everything else
        if not isinstance(json_out[key], dict):
            continue

        # protect against a dict that is not vni
        if json_out[key].get('vni') is None:
            continue
        new_entry = EvpnVniInfo()
        new_entry.adv_all_vni = adv_all_vni
        new_entry.adv_gw_ip = adv_gw_ip

        try:
            new_entry.vni = int(key.encode('ascii', 'ignore'))
            new_entry.origin_ip = json_out[key]['originatorIp'].encode('ascii',
                                                                   'ignore')
            new_entry.rd = json_out[key]['rd'].encode('ascii', 'ignore')
            new_entry.export_rt = json_out[key]['exportRTs']
            new_entry.import_rt = json_out[key]['importRTs']
            new_entry.in_kernel = json_out[key]['inKernel'].encode('ascii',
                                                               'ignore')
            new_entry.save()
        except ValueError:
            # dont crash on invalid json!
            pass

    return retcode
