# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import os
import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import ClagSession
from netq_lib.common.enums import ServiceStatus


def process_clagctl(cmd_output, svc_failed=False, **kwargs):
    """Build clag info for node
    """
    retcode = ServiceStatus.OK

    try:
        json_out = json.loads(cmd_output)
    except ValueError:
        json_out = []
        retcode = ServiceStatus.FAIL

    if 'errorMsg' in json_out:
        if 'Unable to communicate with clagd' in json_out['errorMsg']:
            # Does the clagd file exist ?
            if (os.path.isfile('/etc/default/clagd') and
                    os.access('/etc/default/clagd', os.R_OK)):
                retcode = ServiceStatus.FAIL
                svc_failed = True
            else:
                retcode = ServiceStatus.OK
                svc_failed = False
        else:
            retcode = ServiceStatus.FAIL
            svc_failed = True
        json_out = []

    m_clag_session = ClagSession.memmodel()
    for entry in m_clag_session.query.filter(hostname=HOSTNAME):
        if not entry.peer_state:
            continue
        old_clag_entry = ClagSession()
        if (not json_out or
                json_out['status']['sysMac'] != entry.clag_sysmac):
            if not svc_failed:
                old_clag_entry.copy(entry)
                old_clag_entry.delete()
            else:
                old_clag_entry.peer_state = False
                old_clag_entry.clag_sysmac = entry.clag_sysmac
                old_clag_entry.save()

    if not json_out:
        return retcode

    clag_conflicted_bonds = []
    clag_singly_attached_bonds = []
    clag_proto_down_bonds = []
    dual_attach_bonds = {}

    clag_session = ClagSession()

    clag_session.peer_state = json_out['status']['peerAlive']

    clag_session.peer_if = (
        json_out['status']['peerIf'].encode('ascii', 'ignore')
    )
    clag_session.peer_role = json_out['status']['peerRole'].encode('ascii',
                                                                   'ignore')
    clag_session.role = json_out['status']['ourRole'].encode('ascii', 'ignore')
    clag_session.clag_sysmac = json_out['status']['sysMac'].encode('ascii',
                                                                   'ignore')
    if json_out['status']['backupIp']:
        clag_session.backup_ip = (
            json_out['status']['backupIp'].encode('ascii', 'ignore')
        )
    else:
        clag_session.backup_ip = ''
    clag_session.backup_ip_active = json_out['status']['backupActive']
    clag_session.vxlan_anycast = (
        (json_out['status'].get('vxlanAnycast', '')).encode('ascii', 'ignore'))

    for iface in json_out['clagIntfs']:
        iface = iface.encode('ascii', 'ignore')
        if json_out['clagIntfs'][iface]['status'] == 'dual':
            dual_attach_bonds[iface] = (
                json_out['clagIntfs'][iface]['peerIf'].encode('ascii', 'ignore')
            )
        elif 'conflicts' in json_out['clagIntfs'][iface]:
            clag_conflicted_bonds.append(
                [iface,
                 json_out['clagIntfs'][iface]['conflicts'][0].
                 encode('ascii', 'ignore')]
            )
            retcode = ServiceStatus.ERROR
        elif 'protoDown' in json_out['clagIntfs'][iface]:
            clag_proto_down_bonds.append([
                iface,
                json_out['clagIntfs'][iface]['protoDown'][0].encode('ascii',
                                                                    'ignore')])
            retcode = ServiceStatus.ERROR
        else:
            clag_singly_attached_bonds.append(iface)
            retcode = ServiceStatus.WARNING

    clag_session.conflicted_bonds = clag_conflicted_bonds
    clag_session.single_bonds = clag_singly_attached_bonds
    clag_session.proto_down_bonds = clag_proto_down_bonds
    clag_session.dual_bonds = dual_attach_bonds

    # If the CLAG sysmac changed, we need to delete the older entry.
    m_clag_session = ClagSession.memmodel()
    for entry in m_clag_session.query.filter(hostname=HOSTNAME):
        if entry.clag_sysmac != clag_session.clag_sysmac:
            old_clag_entry = ClagSession()
            old_clag_entry.copy(entry)
            old_clag_entry.delete()
    clag_session.save()

    return retcode
