# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import LnvSession
from netq_lib.common.enums import ServiceStatus


def lnv_handle_bad_output(hostname, role, svc_failed):
    m_sess = LnvSession.memmodel()
    for sess in m_sess.query.filter(hostname=hostname, role=role):
        if sess.state == 'Failed':
            continue
        s = LnvSession()
        if not svc_failed:
            s.copy(sess)
            s.delete()
        else:
            s.state = 'Failed'
            s.role = sess.role
            s.save()


def get_cmd_output(cmd_output, role, svc_failed):
    try:
        json_out = json.loads(cmd_output)
    except ValueError:
        lnv_handle_bad_output(HOSTNAME, role, svc_failed)
        return [], ServiceStatus.FAIL

    if 'errorMsg' in json_out:
        lnv_handle_bad_output(HOSTNAME, role, svc_failed)
        return [], ServiceStatus.FAIL

    return json_out, ServiceStatus.OK


def process_vxrdctl_show(cmd_output, svc_failed=False, **kwargs):
    """Build Lnv vxrd info for node
    """

    json_out, rv = get_cmd_output(cmd_output, 'RD', svc_failed)
    if rv != ServiceStatus.OK:
        return rv

    sess = LnvSession()
    m_sess = LnvSession.memmodel()
    lnv_session = m_sess.query.get(hostname=HOSTNAME, role='RD')
    if lnv_session:
        sess.copy(lnv_session)

    sess.role = 'RD'
    sess.snd_ip = json_out['snd_ip'].encode('ascii', 'ignore')
    sess.repl_mode = 'HER' if json_out['head_rep'] else 'SVC'
    sess.state = 'Failed' if svc_failed else 'Running'
    sess.version = str(json_out['version'])
    sess.save()
    return ServiceStatus.OK


def process_vxrdctl_peers(cmd_output, svc_failed=False, **kwargs):
    """Build Lnv vxrd info for node
    """

    json_out, rv = get_cmd_output(cmd_output, 'RD', svc_failed)
    if rv != ServiceStatus.OK:
        return rv

    sess = LnvSession()
    m_sess = LnvSession.memmodel()
    lnv_session = m_sess.query.get(hostname=HOSTNAME, role='RD')
    if lnv_session:
        sess.copy(lnv_session)
        sess.vnis = json_out
        sess.save()

    return ServiceStatus.OK


def process_vxsndctl_show(cmd_output, svc_failed=False, **kwargs):
    """Build Lnv vxsnd info for node
    """

    json_out, rv = get_cmd_output(cmd_output, 'SND', svc_failed)
    if rv != ServiceStatus.OK:
        return rv

    sess = LnvSession()
    m_sess = LnvSession.memmodel()
    lnv_session = m_sess.query.get(hostname=HOSTNAME, role='SND')
    if lnv_session:
        sess.copy(lnv_session)

    sess.role = 'SND'
    sess.snd_ip = json_out['anycast'].encode('ascii', 'ignore')
    sess.state = 'Failed' if svc_failed else 'Running'
    sess.version = str(json_out['version'])
    sess.snd_peers = json_out['peers']
    sess.save()
    return ServiceStatus.OK


def process_vxsndctl_get_config(cmd_output, svc_failed=False, **kwargs):
    """Build Lnv vxsnd info for node
    """

    json_out, rv = get_cmd_output(cmd_output, 'SND', svc_failed)
    if rv != ServiceStatus.OK:
        return rv

    sess = LnvSession()
    m_sess = LnvSession.memmodel()
    lnv_session = m_sess.query.get(hostname=HOSTNAME, role='SND')
    if lnv_session:
        sess.copy(lnv_session)
        sess.repl_mode = 'SVC' if json_out['enable_vxlan_listen'] else 'HER'
        sess.save()
    return ServiceStatus.OK


def process_vxsndctl_fdb(cmd_output, svc_failed=False, **kwargs):
    """Build Lnv vxsnd info for node
    """

    json_out, rv = get_cmd_output(cmd_output, 'SND', svc_failed)
    if rv != ServiceStatus.OK:
        return rv

    sess = LnvSession()
    m_sess = LnvSession.memmodel()
    lnv_session = m_sess.query.get(hostname=HOSTNAME, role='SND')
    if not lnv_session:
        return ServiceStatus.OK

    fdb = {}
    for vni, rd in json_out.iteritems():
        fdb[vni] = list(set(r[0] for r in rd))

    sess.copy(lnv_session)
    sess.vnis = fdb
    sess.save()
    return ServiceStatus.OK
