# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import ujson as json

from netq_lib.orm.redisdb.models import (
    OS,
    CPU,
    ASIC,
    Board,
    Disk,
    Memory)


def process_platform_info(platform, os_release, decode_syseeprom, lscpu,  # pylint: disable=unused-argument
                          meminfo, platform_json, meminfo_dmi, lsblk):
    os = OS()
    cpu = CPU()
    asic = ASIC()
    board = Board()

    os.name = 'N/A'
    os.version = 'N/A'
    os.version_id = 'N/A'
    cpu.arch = 'N/A'
    cpu.nos = 'N/A'
    cpu.model = 'N/A'
    cpu.max_freq = 'N/A'
    cpu.mem_total = 'N/A'
    asic.vendor = 'N/A'
    asic.model = 'N/A'
    asic.model_id = 'N/A'
    asic.core_bw = 'N/A'
    asic.ports = 'N/A'
    board.vendor = 'N/A'
    board.model = 'N/A'
    board.base_mac = 'N/A'
    board.part_number = 'N/A'
    board.mfg_date = 'N/A'
    board.serial_number = 'N/A'
    board.label_revision = 'N/A'

    if os_release is not None:
        for line in os_release.splitlines():
            if line.startswith('NAME='):
                os.name = line.split('=')[1].strip('\"')
            if line.startswith('VERSION='):
                os.version = line.split('=')[1].strip('\"')
            if line.startswith('VERSION_ID='):
                os.version_id = line.split('=')[1].strip('\"')

    if lscpu is not None:
        for line in lscpu.splitlines():
            if line.startswith('Architecture'):
                cpu.arch = line.split(':')[1].strip()
            if line.startswith('CPU(s)'):
                cpu.nos = line.split(':')[1].strip()
            if line.startswith('Model name'):
                cpu_info = line.split(':')[1].strip()
                model_freq = cpu_info.split('@')
                if len(model_freq) == 1:
                    cpu.model = model_freq[0].replace('  ', '').replace('CPU',
                                                                        '')
                elif len(model_freq) == 2:
                    cpu.model = model_freq[0].replace('  ', '').replace('CPU',
                                                                        '')
                    cpu.max_freq = model_freq[1]

    if decode_syseeprom is not None:
        json_op = json.loads(decode_syseeprom)
        tlvs = json_op['tlv']
        if 'Base MAC Address' in tlvs:
            board.base_mac = tlvs["Base MAC Address"]["value"].encode(
                'ascii', 'ignore').strip(chr(0))
        if 'Part Number' in tlvs:
            board.part_number = tlvs["Part Number"]["value"].encode(
                'ascii', 'ignore').strip(chr(0))
        if 'Manufacture Date' in tlvs:
            board.mfg_date = tlvs["Manufacture Date"]["value"][:10].encode(
                'ascii', 'ignore').strip(chr(0))
        if 'Serial Number' in tlvs:
            board.serial_number = tlvs["Serial Number"]["value"].encode(
                'ascii', 'ignore').strip(chr(0))
        if 'Label Revision' in tlvs:
            board.label_revision = tlvs["Label Revision"]["value"].encode(
                'ascii', 'ignore').strip(chr(0))
        if 'Vendor Name' in tlvs:
            board.vendor = tlvs["Vendor Name"]["value"].encode(
                'ascii', 'ignore').strip(chr(0))
        if 'Product Name' in tlvs:
            board.model = tlvs["Product Name"]["value"].encode(
                'ascii', 'ignore').strip(chr(0))

    if platform_json:
        asic.vendor = platform_json['soc']['vendor'].encode('ascii', 'ignore')
        asic.model = platform_json['soc']['model'].encode('ascii', 'ignore')
        asic.model_id = platform_json['soc']['model_id'].encode('ascii',
                                                                'ignore')
        # CL 3.1.2 and before use /usr/share/cumulus/platform-info/<vendor.json>
        # that don't have the core_bw fields.
        if platform_json['soc'].get('core_bw', None):
            asic.core_bw = '%s%s' % (
                platform_json['soc']['core_bw'].encode('ascii', 'ignore'),
                platform_json['soc']['core_bw_unit'].encode('ascii', 'ignore'))
        asic.ports = platform_json['ports']['layout'].encode('ascii', 'ignore')
        board.model = platform_json['model_name'].encode('ascii', 'ignore')
        board.vendor = platform_json['vendor'].encode('ascii', 'ignore')

    if lsblk is not None:
        for line in lsblk.splitlines():
            disk = Disk()
            disk.name = 'N/A'
            disk.size = 'N/A'
            disk.d_type = 'N/A'
            disk.vendor = 'N/A'
            disk.transport = 'N/A'
            disk.rev = 'N/A'
            disk.model = 'N/A'
            d_str = line.split()
            if len(d_str) > 0:
                disk.name = d_str[0]
            if len(d_str) > 1:
                disk.size = d_str[1]
                try:
                    if disk.size.endswith('M'):
                        disk.size = ('%0.2fG' % (float(
                            disk.size.split('M')[0])/1024))
                except ValueError:
                    pass
            if len(d_str) > 2:
                disk.d_type = d_str[2]
            if len(d_str) > 3:
                disk.vendor = d_str[3]
            if len(d_str) > 4:
                disk.transport = d_str[4]
            if len(d_str) > 5:
                disk.rev = d_str[5]
            if len(d_str) > 6:
                disk.model = d_str[6]
            if len(d_str) > 7:
                disk.model += ' ' + d_str[7]
            disk.save()

    mem_size = []
    mem_locator = []
    mem_b_locator = []
    mem_m_type = []
    mem_serial = []
    mem_mfg = []
    mem_speed = []
    if meminfo_dmi is not None:
        for line in meminfo_dmi.splitlines():
            line = line.strip()
            if line.startswith('Size:'):
                _, val = line.split(':')
                mem_size.append(val)
            elif line.startswith('Bank Locator:'):
                _, val = line.split(':')
                mem_b_locator.append(val)
            elif line.startswith('Locator:'):
                _, val = line.split(':')
                mem_locator.append(val)
            elif line.startswith('Speed:'):
                _, val = line.split(':')
                mem_speed.append(val)
            elif line.startswith('Type:'):
                _, val = line.split(':')
                mem_m_type.append(val)
            elif line.startswith('Serial Number:'):
                _, val = line.split(':')
                mem_serial.append(val)
            elif line.startswith('Manufacturer:'):
                _, val = line.split(':')
                mem_mfg.append(val)
    if (len(mem_size) == 0) and (meminfo is not None):
        for line in meminfo.splitlines():
            if line.startswith('MemTotal'):
                mem_str = float(line.split(':')[1].strip().split()[0])
                mem_size.append('%.2f' % (mem_str / 1024) + 'MB')
                mem_locator.append('N/A')
                mem_b_locator.append('')
                mem_m_type.append('N/A')
                mem_mfg.append('N/A')
                mem_serial.append('N/A')
                mem_speed.append('N/A')
                break

    for i in range(0, len(mem_size)):
        if 'no' in mem_size[i].lower():
            continue
        mem = Memory()
        mem.size = mem_size[i]
        mem.name = mem_locator[i]
        if 'no' not in mem_b_locator[i].lower():
            mem.name += mem_b_locator[i]
        mem.m_type = mem_m_type[i]
        if len(mem_mfg) > i:
            mem.vendor = mem_mfg[i]
            if 'bad' in mem.vendor.lower():
                mem.vendor = 'N/A'
        if len(mem_serial) > i:
            mem.serial_number = mem_serial[i]
            if 'bad' in mem.serial_number.lower():
                mem.serial_number = 'N/A'
        if len(mem_speed) > i:
            mem.speed = mem_speed[i]
        mem.save()

    os.save()
    cpu.save()
    asic.save()
    board.save()

    return 0
