# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

# GENERIC CONFIGURATION HELPERS

from __future__ import absolute_import
from __future__ import print_function

import os
import yaml

from netq_lib.common import utils
from netq_lib.common.enums import DefaultPaths

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

try:
    from cStringIO import StringIO
except:
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

def represent_none(self, _):
    return self.represent_scalar('tag:yaml.org,2002:null', '')

def update_yaml_configuration(conf_dict, path, preamble=None):
    yaml.add_representer(type(None), represent_none)
    try:
        dir_path = os.path.dirname(path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        with open(path, 'w') as fp:
            if preamble:
                fp.write(preamble)
            yaml.dump(conf_dict, fp, default_flow_style=False, default_style='')
    except Exception as ex:
        utils.get_logger('_netq_configuration_').exception(ex.message)
        return 1 # Coincidentally RC_FAIL
    return 0

def get_yaml_configuration(proc_name='', logger=None, paths=None):
    '''aggregate configuration file values, evaluate default paths alphabetically, provided paths exact order'''

    main_conf = dict()
    conf = dict()

    paths = paths
    for conf_file in paths:
        error, msg = load_configuration(conf_file)
        if error:
            if logger:
                logger.error(msg)
            else:
                print(msg)
        else:
            conf = msg

        if not conf: continue

        for key, value in conf.items():
            if not main_conf.get(key):
                main_conf[key] = value
            else:
                if isinstance(main_conf[key],dict) and isinstance(value, dict):
                    #Merge values in original dict with ones from new dict
                    for key2, value2 in value.items():
                        main_conf[key][key2] = value2
                else:
                    #Obliterate original value
                    main_conf[key] = value

    return main_conf

def save_configuration(conf_file, settings):
    try:
        dir_path = os.path.dirname(conf_file)
        if not os.path.exists(dir_path): os.makedirs(dir_path)
        sfile = open(conf_file,'w')
    except IOError:
        pass
    else:
        with sfile:
            try:
                yaml.safe_dump(settings,sfile)
            except yaml.YAMLError:
                pass

def load_configuration(conf_file):
    try:
        sfile = open(conf_file,'r')
    except IOError as ioe:
        return True, ('Unable to open configuration file: %s, error: %s' % (conf_file, str(ioe)))
    else:
        with sfile:
            try:
                return False,yaml.safe_load(sfile)
            except yaml.YAMLError as yex:
                return True,('Unable to parse configuration file: %s, error: %s' % (conf_file, str(yex)))


def get_config(proc_name='', log=False, paths=None):
    '''aggregate configuration file values, evaluate default paths alphabetically, provided paths exact order'''

    # TODO: Move
    logger = utils.get_logger(proc_name) if log else None

    conf = get_yaml_configuration(proc_name, logger, paths=get_paths(paths))

    if logger:
        for hdlr in logger.handlers:
            logger.removeHandler(hdlr)
    del logger

    return conf


def get_paths(paths):
    if paths is None: return __get_configuration_paths()
    return paths if isinstance(paths,list) else [paths]


def __get_configuration_paths():
    conf_files = []

    if os.path.exists(DefaultPaths.CONF_FILE):
        conf_files.append(DefaultPaths.CONF_FILE)

    for root, dirs, files in os.walk(DefaultPaths.CONF_DIR):
        for filename in sorted(files):
            # only look for files with .yml or .yaml extensions
            if filename.endswith('.yml') or filename.endswith('.yaml'):
                conf_files.append(os.path.join(root, filename))

    return conf_files